# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import os
import sys
sys.path.insert(0, os.path.abspath('..'))  # set path for project import

from sarpy import __about__ as parameters  # fetch our relevant project details

# -- Project information -----------------------------------------------------

project = 'sarpy'
version = parameters.__version__  # The full version, including alpha/beta/rc tags
release = parameters.__version__
copyright = parameters.__copyright__
author = parameters.__author__
html_logo = 'nga_logo.jpeg'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc', 'sphinx.ext.coverage', 'sphinxcontrib.napoleon']

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinxdoc'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# control the order of elements shown by autodoc
autodoc_member_order = 'bysource'


"""
This package contains the CRSD schema
"""

__classification__ = 'UNCLASSIFIED'

import pkg_resources


def get_schema_path(version='1.0.0'):
    """
    Location of CRSD schema file.

    Parameters
    ----------
    version : str

    Returns
    -------
    str
        The path to the CRSD schema.
    """

    if version == '1.0.0':
        return pkg_resources.resource_filename('sarpy.io.received.crsd_schema', 'CRSD_schema_V1.0.0_2021_06_12.xsd')
    else:
        raise ValueError('Got unrecognized version {}'.format(version))



"""
Module providing api consistent with other file types for reading tiff files.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Daniel Haverporth")

# It was the original intent to use gdal for the bulk of tiff reading
# Unfortunately, the necessary sarpy functionality can only be obtained by
# gdal_dataset.GetVirtualMemArray(). As of July 2020, this is supported only
# on Linux platforms - unclear what more constraints. So, using gdal to provide
# the reading capability is not feasible at present.

import logging
import os

import numpy
import re
from typing import Tuple

from sarpy.io.general.base import BaseReader, BIPChipper, SarpyIOError


logger = logging.getLogger(__name__)


_BASELINE_TAGS = {
    254: 'NewSubfileType',
    255: 'SubfileType',
    256: 'ImageWidth',
    257: 'ImageLength',
    258: 'BitsPerSample',
    259: 'Compression',
    262: 'PhotometricInterpretation',
    263: 'Thresholding',
    264: 'CellWidth',
    265: 'CellLength',
    266: 'FillOrder',
    270: 'ImageDescription',
    271: 'Make',
    272: 'Model',
    273: 'StripOffsets',
    274: 'Orientation',
    277: 'SamplesPerPixel',
    278: 'RowsPerStrip',
    279: 'StripByteCounts',
    280: 'MinSampleValue',
    281: 'MaxSampleValue',
    282: 'XResolution',
    283: 'YResolution',
    284: 'PlanarConfiguration',
    288: 'FreeOffsets',
    289: 'FreeByteCounts',
    290: 'GrayResponseUnit',
    291: 'GrayResponseCurve',
    296: 'ResolutionUnit',
    305: 'Software',
    306: 'DateTime',
    315: 'Artist',
    316: 'HostComputer',
    320: 'ColorMap',
    338: 'ExtraSamples',
    33432: 'Copyright',
}
_EXTENSION_TAGS = {
    269: 'DocumentName',
    285: 'PageName',
    286: 'XPosition',
    287: 'YPosition',
    292: 'T4Options',
    293: 'T6Options',
    297: 'PageNumber',
    301: 'TransferDunction',
    317: 'Predictor',
    318: 'WhitePoint',
    319: 'PrimaryChromaticities',
    321: 'HalftoneHints',
    322: 'TileWidth',
    323: 'TileLength',
    324: 'TileOffsets',
    325: 'TileByteCounts',
    326: 'BadFaxLines',
    327: 'CleanFaxData',
    328: 'ConsecutiveBadFaxLines',
    330: 'SubIFDs',
    332: 'InkSet',
    333: 'InkNames',
    334: 'NumberOfInks',
    336: 'DotRange',
    337: 'TargetPrinter',
    339: 'SampleFormat',
    340: 'SMinSampleValue',
    341: 'SMaxSampleValue',
    342: 'TransferRange',
    343: 'ClipPath',
    344: 'XClipPathUnits',
    345: 'YClipPathUnits',
    346: 'Indexed',
    347: 'JPEGTables',
    351: 'OPIProxy',
    400: 'GlobalParametersIFD',
    401: 'ProfileType',
    402: 'FaxProfile',
    403: 'CodingMethods',
    404: 'VersionYear',
    405: 'ModeNumber',
    433: 'Decode',
    434: 'DefaultImageColor',
    512: 'JPEGProc',
    513: 'JPEGInterchangeFormat',
    514: 'JPEGInterchangeFormatLength',
    515: 'JPEGRestartInterval',
    517: 'JPEGLosslessPredictors',
    518: 'JPEGPointTransforms',
    519: 'JPEGQTables',
    520: 'JPEGDCTables',
    521: 'JPEGACTables',
    529: 'YCbCrCoefficients',
    530: 'YCbCrSubSampling',
    531: 'YCbCrPositioning',
    532: 'ReferenceBlackWhite',
    559: 'StripRowCounts',
    700: 'XMP',
    32781: 'ImageID',
    34732: 'ImageLayer',
}
_GEOTIFF_TAGS = {
    33550: 'ModelPixelScaleTag',
    33922: 'ModelTiePointTag',
    34264: 'ModelTransformationTag',
    34735: 'GeoKeyDirectoryTag',
    34736: 'GeoDoubleParamsTag',
    34737: 'GeoAsciiParamsTag',
}

########
# base expected functionality for a module with an implemented Reader


def is_a(file_name):
    """
    Tests whether a given file_name corresponds to a tiff file. Returns a
    tiff reader instance, if so.

    Parameters
    ----------
    file_name : str
        the file_name to check

    Returns
    -------
    TiffReader|None
        `TiffReader` instance if tiff file, `None` otherwise
    """

    try:
        tiff_details = TiffDetails(file_name)
        logger.info('File {} is determined to be a tiff file.'.format(file_name))
        return TiffReader(tiff_details)
    except SarpyIOError:
        # we don't want to catch parsing errors, for now
        return None


##########


class TiffDetails(object):
    """
    For checking tiff metadata, and parsing in the event we are not using GDAL
    """

    __slots__ = ('_file_name', '_endian', '_magic_number', '_tags')
    _DTYPES = {i+1: entry for i, entry in enumerate(
        ['u1', 'a', 'u2', 'u4', 'u4',
         'i1', 'u1', 'i2', 'i4', 'i4',
         'f4', 'f8', 'u4', None, None,
         'u8', 'i8', 'u8'])}
    _SIZES = numpy.array(
        [1, 1, 2, 4, 8,
         1, 1, 2, 4, 8,
         4, 8, 4, 0, 0,
         8, 8, 8], dtype=numpy.int64)
    # no definition for entries for 14 & 15

    def __init__(self, file_name):
        """

        Parameters
        ----------
        file_name : str
        """

        if not (isinstance(file_name, str) and os.path.isfile(file_name)):
            raise SarpyIOError('Not a TIFF file.')

        with open(file_name, 'rb') as fi:
            # Try to read the basic tiff header
            try:
                fi_endian = fi.read(2).decode('utf-8')
            except Exception as e:
                raise SarpyIOError('Failed decoding the 2 character tiff header with error\n\t{}'.format(e))

            if fi_endian == 'II':
                self._endian = '<'
            elif fi_endian == 'MM':
                self._endian = '>'
            else:
                raise SarpyIOError('Invalid tiff endian string {}'.format(fi_endian))
            # check the magic number
            self._magic_number = numpy.fromfile(fi, dtype='{}i2'.format(self._endian), count=1)[0]
            if self._magic_number not in [42, 43]:
                raise SarpyIOError('Not a valid tiff file, got magic number {}'.format(self._magic_number))

            if self._magic_number == 43:
                rhead = numpy.fromfile(fi, dtype='{}i2'.format(self._endian), count=2)
                if rhead[0] != 8:
                    raise SarpyIOError('Not a valid bigtiff. The offset size is given as {}'.format(rhead[0]))
                if rhead[1] != 0:
                    raise SarpyIOError('Not a valid bigtiff. The reserved entry of '
                                  'the header is given as {} != 0'.format(rhead[1]))
        self._file_name = file_name
        self._tags = None

    @property
    def file_name(self):
        """
        str: READ ONLY. The file name.
        """

        return self._file_name

    @property
    def endian(self):
        """
        str: READ ONLY. The numpy dtype style ``('>' = big, '<' = little)`` endian string for the tiff file.
        """

        return self._endian

    @property
    def tags(self):
        """
        None|dict: READ ONLY. the tiff tags dictionary, provided that func:`parse_tags` has been called.
        This dictionary is of the form ``{<tag name> : numpy.ndarray value}``, even for
        those tags containing only a single entry (i.e. `count=1`).
        """
        if self._tags is None:
            self.parse_tags()
        return self._tags

    def parse_tags(self):
        """
        Parse the tags from the file, if desired. This sets the `tags` attribute.

        Returns
        -------
        None
        """

        if self._magic_number == 42:
            type_dtype = numpy.dtype('{}u2'.format(self._endian))
            count_dtype = numpy.dtype('{}u2'.format(self._endian))
            offset_dtype = numpy.dtype('{}u4'.format(self._endian))
            offset_size = 4
        elif self._magic_number == 43:
            type_dtype = numpy.dtype('{}u2'.format(self._endian))
            count_dtype = numpy.dtype('{}i8'.format(self._endian))
            offset_dtype = numpy.dtype('{}i8'.format(self._endian))
            offset_size = 8
        else:
            raise ValueError('Unrecognized magic number {}'.format(self._magic_number))

        with open(self._file_name, 'rb') as fi:
            # skip the basic header
            fi.seek(offset_size, os.SEEK_SET)
            # extract the tags information
            tags = {}
            self._parse_ifd(fi, tags, type_dtype, count_dtype, offset_dtype, offset_size)
        self._tags = tags

    def _read_tag(self, fi, tiff_type, num_tag, count):
        """
        Parse the specific tag information.

        Parameters
        ----------
        fi
            The file type object
        tiff_type : int
            The tag data type identifier value
        num_tag : int
            The numeric tag identifier
        count : int
            The number of such tags

        Returns
        -------
        dict
        """

        # find which tags we belong to
        if num_tag in _BASELINE_TAGS:
            ext = 'BaselineTag'
            name = _BASELINE_TAGS[num_tag]
        elif num_tag in _EXTENSION_TAGS:
            ext = 'ExtensionTag'
            name = _EXTENSION_TAGS[num_tag]
        elif num_tag in _GEOTIFF_TAGS:
            ext = 'GeoTiffTag'
            name = _GEOTIFF_TAGS[num_tag]
        else:
            ext, name = None, None
        # Now extract from file based on type number
        dtype = self._DTYPES.get(int(tiff_type), None)
        if dtype is None:
            logger.warning(
                'Failed to extract tiff data type {},\n\t'
                'for {} - {}'.format(tiff_type, ext, name))
            return {'Value': None, 'Name': name, 'Extension': ext}
        if tiff_type == 2:  # ascii field - read directly and decode?
            val = fi.read(count)  # this will be a string for python 2, and we decode for python 3
            if not isinstance(val, str):
                val = val.decode('utf-8')
            # eliminate the null characters
            val = re.sub('\x00', '', val)
        elif tiff_type in [5, 10]:  # unsigned or signed rational
            val = numpy.fromfile(fi, dtype='{}{}'.format(self._endian, dtype), count=numpy.int64(2*count)).reshape((-1, 2))
        else:
            val = numpy.fromfile(fi, dtype='{}{}'.format(self._endian, dtype), count=count)
            if count == 1:
                val = val[0]
        return {'Value': val, 'Name': name, 'Extension': ext}

    def _parse_ifd(self, fi, tags, type_dtype, count_dtype, offset_dtype, offset_size):
        """
        Recursively parses the tag data and populates a provided dictionary
        Parameters
        ----------
        fi
            The file type object
        tags : dict
            The tag data dictionary being populated
        type_dtype : str|numpy.dtype
            The data type for the data element - note that endian-ness is included
        count_dtype : str|numpy.dtype
            The data type for the number of directories - note that endian-ness is included
        offset_dtype : str|numpy.dtype
            The data type for the offset - note that endian-ness is included
        offset_size : int
            The size of the offset
        Returns
        -------
        None
        """

        nifd = numpy.fromfile(fi, dtype=offset_dtype, count=1)[0]
        if nifd == 0:
            return  # termination criterion

        fi.seek(nifd)
        num_entries = numpy.fromfile(fi, dtype=count_dtype, count=1)[0]
        for entry in range(int(num_entries)):
            num_tag, tiff_type = numpy.fromfile(fi, dtype=type_dtype, count=2)
            count = numpy.fromfile(fi, dtype=offset_dtype, count=1)[0]
            total_size = self._SIZES[tiff_type-1]*count
            if total_size <= offset_size:
                save_ptr = fi.tell() + offset_size  # we should advance past the entire block
                value = self._read_tag(fi, tiff_type, num_tag, count)
                fi.seek(save_ptr)
            else:
                offset = numpy.fromfile(fi, dtype=offset_dtype, count=1)[0]
                save_ptr = fi.tell()  # save our current spot
                fi.seek(offset)  # get to the offset location
                value = self._read_tag(fi, tiff_type, num_tag, count)  # read the tag value
                fi.seek(save_ptr)  # return to our location
            tags[value['Name']] = value['Value']
        self._parse_ifd(fi, tags, type_dtype, count_dtype, offset_dtype, offset_size)  # recurse

    def check_compression(self):
        """
        Check the Compression tag, and verify uncompressed.

        Returns
        -------
        None
        """

        if self.tags['Compression'] != 1:
            raise ValueError(
                'The file {} indicates some kind of tiff compression, and the sarpy API requirements '
                'do not presently support reading of compressed tiff files. Consider using gdal to '
                'translate this tiff to an uncompressed file via the commmand\n\t'
                '"gdal_translate -co TILED=no <input_file> <output_file>"')

    def check_tiled(self):
        """
        Check if the tiff file is tiled.

        Returns
        -------
        None
        """

        if 'TileLength' in self.tags or 'TileWidth' in self.tags:
            raise ValueError(
                'The file {} indicates that this is a tiled file, and the sarpy API requirements '
                'do not presently support reading of tiled tiff files. Consider using gdal to '
                'translate this tiff to a flat file via the commmand\n\t'
                '"gdal_translate -co TILED=no <input_file> <output_file>"')


class NativeTiffChipper(BIPChipper):
    """
    Direct reading of data from tiff file, failing if compression is present
    """

    __slots__ = ('_tiff_details', )
    _SAMPLE_FORMATS = {
        1: 'u', 2: 'i', 3: 'f', 5: 'i', 6: 'f'}  # 5 and 6 are complex int/float

    def __init__(self, tiff_details, symmetry=(False, False, True)):
        """

        Parameters
        ----------
        tiff_details : TiffDetails
        symmetry : Tuple[bool]
        """

        if isinstance(tiff_details, str):
            tiff_details = TiffDetails(tiff_details)
        if not isinstance(tiff_details, TiffDetails):
            raise TypeError('NativeTiffChipper input argument must be a filename '
                            'or TiffDetails object.')

        tiff_details.check_compression()
        tiff_details.check_tiled()

        self._tiff_details = tiff_details
        if isinstance(tiff_details.tags['SampleFormat'], numpy.ndarray):
            samp_form = tiff_details.tags['SampleFormat'][0]
        else:
            samp_form = tiff_details.tags['SampleFormat']
        if samp_form not in self._SAMPLE_FORMATS:
            raise ValueError('Invalid sample format {}'.format(samp_form))
        if isinstance(tiff_details.tags['BitsPerSample'], numpy.ndarray):
            bits_per_sample = tiff_details.tags['BitsPerSample'][0]
        else:
            bits_per_sample = tiff_details.tags['BitsPerSample']

        raw_bands = int(tiff_details.tags['SamplesPerPixel'])

        if samp_form in [5, 6]:
            transform_data = 'COMPLEX'
            output_bands = int(raw_bands)
            raw_bands *= 2
            bits_per_sample /= 2
            output_dtype = 'complex64'
        elif raw_bands == 2:
            # NB: this is heavily skewed towards SAR and obviously not general
            transform_data = 'COMPLEX'
            output_dtype = 'complex64'
            output_bands = 1
        else:
            transform_data = None
            output_bands = raw_bands
            output_dtype = None

        data_size = (int(tiff_details.tags['ImageLength']), int(tiff_details.tags['ImageWidth']))
        raw_dtype = numpy.dtype('{0:s}{1:s}{2:d}'.format(
            self._tiff_details.endian, self._SAMPLE_FORMATS[samp_form], int(bits_per_sample/8)))

        if output_dtype is None:
            output_dtype = raw_dtype
        data_offset = int(tiff_details.tags['StripOffsets'][0])

        super(NativeTiffChipper, self).__init__(
            tiff_details.file_name, raw_dtype, data_size, raw_bands, output_bands, output_dtype,
            symmetry=symmetry, transform_data=transform_data, data_offset=data_offset)


class TiffReader(BaseReader):
    __slots__ = ('_tiff_details', )
    _DEFAULT_SYMMETRY = (False, False, False)

    def __init__(self, tiff_details, symmetry=None):
        """

        Parameters
        ----------
        tiff_details : TiffDetails
        symmetry : Tuple[bool]
        """

        if isinstance(tiff_details, str):
            tiff_details = TiffDetails(tiff_details)
        if not isinstance(tiff_details, TiffDetails):
            raise TypeError('TiffReader input argument must be a filename '
                            'or TiffDetails object.')

        self._tiff_details = tiff_details
        if symmetry is None:
            symmetry = self._DEFAULT_SYMMETRY

        chipper = NativeTiffChipper(tiff_details, symmetry=symmetry)
        super(TiffReader, self).__init__(chipper, reader_type="OTHER")

    @property
    def tiff_details(self):
        # type: () -> TiffDetails
        """
        TiffDetails: The tiff details object.
        """

        return self._tiff_details

    @property
    def file_name(self):
        return self.tiff_details.file_name



"""
Module for reading and interpreting the standard change detection result files.

**This requires optional dependency `pyproj`, and very likely requires `Pillow`
for dealing with a NITF with compressed image segments.**
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
import os
import json
from typing import Union, Dict, Tuple

import numpy

from sarpy.io.general.base import AggregateReader, SarpyIOError
from sarpy.io.general.nitf import NITFReader
from sarpy.geometry.geometry_elements import FeatureCollection, Feature, Geometry

logger = logging.getLogger(__name__)

try:
    # noinspection PyPackageRequirements
    import pyproj
    # NB: this very likely requires PIL (compressed images),
    # which is handled in sarpy.io.general.nitf
except ImportError:
    logger.error(
        'Optional dependency pyproj is required for interpretation '
        'of standard change detection products.')
    pyproj = None


def _validate_coords(coords):
    """
    Validate any coordinate arrays.

    Parameters
    ----------
    coords : numpy.ndarray|list|tuple

    Returns
    -------
    (numpy.ndarray, tuple)
    """

    if not isinstance(coords, numpy.ndarray):
        coords = numpy.array(coords, dtype='float64')

    orig_shape = coords.shape

    if coords.shape[-1] < 2:
        raise ValueError(
            'The coords array must must have final dimension length at least 2. '
            'We have shape = {}'.format(coords.shape))

    if len(coords.shape) == 1:
        coords = numpy.reshape(coords, (1, -1))
    if coords.shape[-1] > 2:
        return coords[:, :2], orig_shape[:-1] + (2, )
    else:
        return coords, orig_shape


def _get_projection(corner_string, hemisphere, data_size):
    """
    Gets the projection method for [lon, lat] -> [row, col].

    Parameters
    ----------
    corner_string : str
    hemisphere : str
    data_size : tuple

    Returns
    -------
    callable
    """

    if pyproj is None:
        raise ValueError('This requires dependency pyproj.')

    # make sure that hemisphere is sensible
    if hemisphere not in ['S', 'N']:
        raise ValueError('hemisphere must be one of "N" or "S", got {}'.format(hemisphere))

    # split into the respective corner parts
    corners_strings = [corner_string[start:stop] for start, stop in zip(range(0, 59, 15), range(15, 74, 15))]
    # parse the utm zone
    utm_zone = corners_strings[0][:2]
    for entry in corners_strings:
        if entry[:2] != utm_zone:
            raise ValueError('Got incompatible corner UTM zones {} and {}'.format(utm_zone, entry[:2]))
    utm_zone = int(utm_zone)
    # parse the corner strings into utm coordinates
    utms = [(float(frag[2:8].strip()), float(frag[8:].strip())) for frag in corners_strings]
    utms = numpy.array(utms, dtype='float64')

    rows, cols = data_size
    col_vector = (utms[1, :] - utms[0, :])/(cols - 1)
    row_vector = (utms[3, :] - utms[0, :])/(rows - 1)
    if numpy.abs(row_vector.dot(col_vector)) > 1e-6:
        raise ValueError('This does not appear to be an ortho-rectified image.')
    test_corner = utms[0, :] + (rows -1)*row_vector + (cols-1)*col_vector
    if numpy.any(numpy.abs(test_corner - utms[2, :]) > 1):
        raise ValueError('This does not appear to be an ortho-rectified image.')
    # define our projection - this is for projecting lon,lat to our UTM coords (and vice versa)
    the_proj = pyproj.Proj(proj='utm',zone=utm_zone, south=(hemisphere == 'S'), ellps='WGS84')
    # account for row/column spacing for pixel conversion
    row_vector /= numpy.sum(row_vector*row_vector)
    col_vector /= numpy.sum(col_vector*col_vector)

    def projection_method(lon_lat):
        # project lon/lat to utm
        lon_lat, o_shape = _validate_coords(lon_lat)
        lon_lat = numpy.reshape(lon_lat, (-1, 2))
        xy = numpy.zeros(lon_lat.shape, dtype='float64')
        xy[:, 0], xy[:, 1] = the_proj(lon_lat[:, 0], lon_lat[:, 1])
        # recenter based on the first corner
        xy -= utms[0, :]
        # convert to pixel
        pixel_xy = numpy.zeros(lon_lat.shape, dtype='float64')
        pixel_xy[:, 0] = xy.dot(row_vector)
        pixel_xy[:, 1] = xy.dot(col_vector)
        return pixel_xy
    return projection_method


class ChangeDetectionDetails(object):
    """
    This is a helper class for interpreting the standard change detection results.
    """

    __slots__ = (
        '_file_names', '_features', '_head_feature')

    def __init__(self, file_name):
        """

        Parameters
        ----------
        file_name : str
        """

        self._file_names = {}
        self._features = None
        self._head_feature = None

        if not isinstance(file_name, str):
            raise SarpyIOError('file_name is required to be a string path name.')
        if not os.path.isfile(file_name):
            raise SarpyIOError('file_name = {} is not a valid existing file'.format(file_name))

        root_dir, fname = os.path.split(file_name)
        fstem, fext = os.path.splitext(fname)
        if fext not in ['.ntf', '.nitf', '.json']:
            raise SarpyIOError('file_name is expected to have extension .ntf, .nitf, .json')
        if not (fstem.endswith('_C') or fstem.endswith('_M') or fstem.endswith('_R') or fstem.endswith('_V')):
            raise SarpyIOError(
                'file_name is expected to follow naming pattern XXX_C.ntf, XXX_M.ntf, XXX_R.ntf, '
                'or XXX_V.json')
        fstem = fstem[:-2]
        for part in ['C', 'M', 'R']:
            for ext in ['.ntf', '.nitf']:
                fil = os.path.join(root_dir, '{}_{}{}'.format(fstem, part, ext))
                if os.path.isfile(fil):
                    self._file_names[part] = fil
            if part not in self._file_names:
                raise SarpyIOError('Change detection file part {} not found'.format(part))
        fil = os.path.join(root_dir, '{}_V.json'.format(fstem))
        if os.path.isfile(fil):
            self._file_names['V'] = fil
        else:
            logger.warning('Change detection file part V (i.e. json metadata) not found.')
        self._set_features()

    @property
    def file_names(self):
        # type: () -> Dict[str, str]
        """
        Dict[str, str] : The file names.
        """

        return self._file_names

    def _set_features(self):
        if self._features is not None:
            return

        if 'V' in self._file_names:
            the_file = self._file_names['V']
            try:
                with open(the_file, 'r') as fi:
                    the_dict = json.load(fi)
                    self._features = FeatureCollection.from_dict(the_dict)
            except Exception as e:
                logger.error(
                    'Failed decoding change detection json file {}\n\twith error {}.\n\t'
                    'Skipping json definition.'.format(the_file, e))
                del self._file_names['V']  # drop this non-functional file to avoid repeating
        if self._features is not None:
            self._head_feature = self._features[0]

    @property
    def features(self):
        # type: () -> Union[FeatureCollection, None]
        """
        None|FeatureCollection: The feature list, if it is defined.
        """
        return self._features

    @property
    def head_feature(self):
        # type: () -> Union[Feature, None]
        """
        None|Feature: The main metadata Feature, which is at the head of the feature list.
        """

        return self._head_feature


class ChangeDetectionReader(AggregateReader):
    """
    An aggregate reader for the standard change detection scenario package files.
    The order of the aggregation is given by `C` (index 0), `M` (index 1), `R` (index 2).
    """

    __slots__ = ('_change_details', '_pixel_geometries')

    def __init__(self, change_details):
        """

        Parameters
        ----------
        change_details : str|ChangeDetectionDetails
        """

        self._pixel_geometries = {}  # type: Dict[str, Geometry]

        if isinstance(change_details, str):
            change_details = ChangeDetectionDetails(change_details)
        if not isinstance(change_details, ChangeDetectionDetails):
            raise TypeError(
                'change_details is required to be the file name of one of the change '
                'detection package files (XXX_C.ntf, XXX_M.ntf, XXX_R.ntf, or XXX_V.json) '
                'or a ChangeDetectionDetails instance. Got type {}'.format(change_details))
        self._change_details = change_details
        readers = (NITFReader(change_details.file_names['C']),
                   NITFReader(change_details.file_names['M']),
                   NITFReader(change_details.file_names['R']),)
        super(ChangeDetectionReader, self).__init__(readers, reader_type="OTHER")
        self._set_pixel_geometries()

    @property
    def readers(self):
        # type: () -> Tuple[NITFReader]
        return self._readers

    @property
    def features(self):
        # type: () -> Union[FeatureCollection, None]
        """
        None|FeatureCollection: The feature list, if it is defined.
        """

        return self._change_details.features

    @property
    def head_feature(self):
        # type: () -> Union[Feature, None]
        """
        None|Feature: The main metadata Feature, which is at the head of the feature list.
        """

        return self._change_details.head_feature

    def _extract_geolocation_details(self):
        """
        Extract the projection method `[lon, lat, hae] -> [pixel row, pixel col]`
        for the reader(s).

        Returns
        -------
        callable
        """

        corner_string = None
        hemisphere = None
        data_size = None
        for reader in self.readers:
            if len(reader.nitf_details.img_headers) != 1:
                raise ValueError(
                    'Each reader is expected to have a single image segment, while reader for file\n'
                    '{}\n has {} segments.'.format(reader.file_name, len(reader.nitf_details.img_headers)))
            img_head = reader.nitf_details.img_headers[0]
            if corner_string is None:
                corner_string = img_head.IGEOLO
                hemisphere = img_head.ICORDS
                data_size = reader.get_data_size_as_tuple()[0]
            else:
                if corner_string != img_head.IGEOLO:
                    raise ValueError(
                        'Got two different IGEOLO entries {} and {}'.format(corner_string, img_head.IGEOLO))
                if hemisphere != img_head.ICORDS:
                    raise ValueError(
                        'Got two different ICORDS entires {} and {}'.format(hemisphere, img_head.ICORDS))
                if data_size != reader.get_data_size_as_tuple()[0]:
                    raise ValueError(
                        'Got two different data sizes {} and {}'.format(data_size, reader.get_data_size_as_tuple()[0]))
        if hemisphere not in ['N', 'S']:
            raise ValueError('Got unexpected ICORDS {}'.format(hemisphere))
        return _get_projection(corner_string, hemisphere, data_size)

    def _set_pixel_geometries(self):
        """
        Sets the pixel geometries.

        Returns
        -------
        None
        """

        feats = self.features
        if feats is None:
            return

        proj_method = self._extract_geolocation_details()
        for feat in feats.features:
            if feat.geometry is None or feat.uid in self._pixel_geometries:
                continue
            self._pixel_geometries[feat.uid] = feat.geometry.apply_projection(proj_method)



"""
Convert from complex SAR image format to SICD format.

For a basic help on the command-line, check

>>> python -m sarpy.utils.convert_to_sicd --help

"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import argparse
import logging

import sarpy
from sarpy.io.complex.converter import conversion_utility


def convert(input_file, output_dir, preserve_nitf_information=False):
    """

    Parameters
    ----------
    input_file : str
        Path to the input file.
    output_dir : str
        Output directory path.
    preserve_nitf_information : bool
        Try to preserve NITF information? This only applies in the case that the
        file being read is actually a NITF file.
    """

    conversion_utility(input_file, output_dir, preserve_nitf_information=preserve_nitf_information)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Convert to SICD format.",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        'input_file', metavar='input_file',
        help='Path input data file, or directory for radarsat, RCM, or sentinel.\n'
             '* For radarsat or RCM, this can be the product.xml file, or parent directory\n'
             '  of product.xml or metadata/product.xml.\n'
             '* For sentinel, this can be the manifest.safe file, or parent directory of\n'
             '  manifest.safe.\n')
    parser.add_argument(
        'output_directory', metavar='output_directory',
        help='Path to the output directory. This directory MUST exist.\n'
             '* Depending on the input details, multiple SICD files may be produced.\n'
             '* The name for the ouput file(s) will be chosen based on CoreName and\n '
             '  transmit/collect polarization.\n')
    parser.add_argument(
        '-p', '--preserve', action='store_true',
        help='Try to preserve any NITF information?\n'
             'This only applies in the event that the file being read is a NITF')
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='Verbose (level="INFO") logging?')

    args = parser.parse_args()
    level = 'INFO' if args.verbose else 'WARNING'
    logging.basicConfig(level=level)
    logger = logging.getLogger('sarpy')
    logger.setLevel(level)

    convert(args.input_file, args.output_directory, preserve_nitf_information=args.preserve)



"""
This package contains the CPHD schema
"""

__classification__ = 'UNCLASSIFIED'

import pkg_resources


def get_schema_path(version='1.0.1'):
    """
    Location of CPHD schema file.

    Returns
    -------
    str
        The path to the CPHD schema.
    """

    if version == '1.0.1':
        return pkg_resources.resource_filename('sarpy.io.phase_history.cphd_schema', 'CPHD_schema_V1.0.1_2018_05_21.xsd')
    else:
        raise ValueError('Got unrecognized version {}'.format(version))



__classification__ = 'UNCLASSIFIED'
__author__ = "Thomas McCullough"

import os
import re
import logging
from typing import List


logger = logging.getLogger('validation')

_the_directory = os.path.split(__file__)[0]
urn_mapping = {
    'urn:SIDD:1.0.0': {
        'ism_urn': 'urn:us:gov:ic:ism',
        'sfa_urn': 'urn:SFA:1.2.0',
        'sicommon_urn': 'urn:SICommon:0.1',
        'version': '1.0',
        'release': '1.0.0',
        'date': '2011-08-31T00:00:00Z',
        'schema': os.path.join(_the_directory, 'version1', 'SIDD_schema_V1.0.0_2011_08_31.xsd')},
    'urn:SIDD:2.0.0': {
        'ism_urn': 'urn:us:gov:ic:ism:13',
        'sfa_urn': 'urn:SFA:1.2.0',
        'sicommon_urn': 'urn:SICommon:1.0',
        'version': '2.0',
        'release': '2.0.0',
        'date': '2019-05-31T00:00:00Z',
        'schema': os.path.join(_the_directory, 'version2', 'SIDD_schema_V2.0.0_2019_05_31.xsd')},
}
_SIDD_SPECIFICATION_IDENTIFIER = 'SIDD Volume 1 Design & Implementation Description Document'


def get_specification_identifier():
    """
    Get the SIDD specification identifier string.

    Returns
    -------
    str
    """

    return _SIDD_SPECIFICATION_IDENTIFIER


def check_urn(urn_string):
    """
    Checks that the urn string follows the correct pattern. This raises an
    exception for a poorly formed or unmapped SIDD urn.

    Parameters
    ----------
    urn_string : str
    """

    if not isinstance(urn_string, str):
        raise TypeError(
            'Expected a urn input of string type, got type {}'.format(type(urn_string)))

    the_match = re.match(r'^urn:SIDD:\d.\d.\d$', urn_string)
    if the_match is None:
        raise ValueError(
            'Input provided as `{}`,\nbut should be of the form '
            '`urn:SIDD:<major>.<minor>.<release>'.format(urn_string))


def get_urn_details(urn_string):
    """
    Gets the associated details for the given SIDD urn, or raise an exception for
    poorly formatted or unrecognized urn.

    Parameters
    ----------
    urn_string

    Returns
    -------
    dict
    """

    check_urn(urn_string)
    out = urn_mapping.get(urn_string, None)

    if out is None:
        raise KeyError(
            'Got correctly formatted, but unmapped SIDD urn {}.'.format(urn_string))
    return out


def get_schema_path(the_urn):
    """
    Gets the path to the proper schema file for the given SIDD urn.

    Parameters
    ----------
    the_urn : str

    Returns
    -------
    str
    """

    result = get_urn_details(the_urn)
    return result['schema']


def get_versions():
    """
    Gets a list of recognized SIDD urn.

    Returns
    -------
    List[str]
    """

    return list(sorted(urn_mapping.keys()))


def validate_xml_ns(xml_ns, ns_key='default'):
    """
    Validate the parsed SIDD xml namespace dictionary. This is expected to
    accompany the use of :func:`sarpy.io.general.utils.parse_xml_from_string`.

    Parameters
    ----------
    xml_ns : dict
        The xml namespace dictionary.
    ns_key : str
        The main SIDD element or default namespace.

    Returns
    -------
    bool
    """

    def validate_ism_urn():
        if 'ism' not in xml_ns:
            the_val = None
            for key in xml_ns:
                val = xml_ns[key]
                if val.lower().startswith('urn:us:gov:ic:ism'):
                    the_val = val
            xml_ns['ism'] = the_val

        valid = True
        if 'ism' not in xml_ns:
            logger.error('SIDD: No `ism` namespace defined.')
            valid = False
        elif xml_ns['ism'] != details['ism_urn']:
            logger.error(
                'SIDD: SIDD {} `ISM` namespace urn is expected to be "{}", but we got "{}".\n\t'
                'Differences in standard may lead to deserialization and/or '
                'validation errors.'.format(sidd_urn, details['ism_urn'], xml_ns['ism']))
            valid = False
        return valid

    def validate_sfa_urn():
        if 'sfa' not in xml_ns:
            the_val = None
            for key in xml_ns:
                val = xml_ns[key]
                if val.lower().startswith('urn:sfa:'):
                    the_val = val
            xml_ns['sfa'] = the_val

        valid = True
        if 'ism' not in xml_ns:
            logger.error('SIDD: No `sfa` namespace defined.')
            valid = False
        elif xml_ns['sfa'] != details['sfa_urn']:
            logger.error(
                'SIDD: SIDD {} `SFA` namespace urn is expected to be "{}", but we got "{}".\n\t'
                'Differences in standard may lead to deserialization and/or '
                'validation errors.'.format(sidd_urn, details['sfa_urn'], xml_ns['sfa']))
            valid = False
        return valid

    def validate_sicommon_urn():
        if 'sicommon' not in xml_ns:
            the_val = None
            for key in xml_ns:
                val = xml_ns[key]
                if val.lower().startswith('urn:sicommon:'):
                    the_val = val
            xml_ns['sicommon'] = the_val

        valid = True
        if 'sicommon' not in xml_ns:
            logger.error('SIDD: No `sicommon` namespace defined.')
            valid = False
        elif xml_ns['sicommon'] != details['sicommon_urn']:
            logger.error(
                'SIDD: SIDD {} `SICommon` namespace urn is expected to be "{}", but we got "{}".\n\t'
                'Differences in standard may lead to deserialization and/or '
                'validation errors.'.format(sidd_urn, details['sicommon_urn'], xml_ns['sicommon']))
            valid = False
        return valid

    if not isinstance(xml_ns, dict):
        return ValueError('xml_ns must be a dictionary for SIDD interpretation.')

    if ns_key not in xml_ns:
        raise ValueError('ns_key must be a key in xml_ns.')

    sidd_urn = xml_ns[ns_key]

    try:
        details = get_urn_details(sidd_urn)
    except KeyError:
        logger.error('Got unmapped sidd urn `{}`'.format(sidd_urn))
        return False

    valid_ns = validate_ism_urn()
    valid_ns &= validate_sfa_urn()
    valid_ns &= validate_sicommon_urn()
    return valid_ns



"""
Common functionality for converting metadata
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
from typing import Iterator, Tuple, List

import numpy
from numpy.polynomial import polynomial
from scipy.stats import scoreatpercentile
from scipy.linalg import lstsq

from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd_elements.blocks import Poly2DType

logger = logging.getLogger(__name__)


def two_dim_poly_fit(x, y, z, x_order=2, y_order=2, x_scale=1, y_scale=1, rcond=None):
    """
    Perform fit of data to two dimensional polynomial.

    Parameters
    ----------
    x : numpy.ndarray
        the x data
    y : numpy.ndarray
        the y data
    z : numpy.ndarray
        the z data
    x_order : int
        the order for x
    y_order : int
        the order for y
    x_scale : float
        In order to help the fitting problem to become better conditioned, the independent
        variables can be scaled, the fit performed, and then the solution rescaled.
    y_scale : float
    rcond : None|float
        passed through to :func:`scipy.linalg.lstsq` (as `cond`).

    Returns
    -------
    numpy.ndarray
        the coefficient array
    """

    if not isinstance(x, numpy.ndarray) or not isinstance(y, numpy.ndarray) or not isinstance(z, numpy.ndarray):
        raise TypeError('x, y, z must be numpy arrays')
    if (x.size != z.size) or (y.size != z.size):
        raise ValueError('x, y, z must have the same cardinality size.')

    x = x.flatten()*x_scale
    y = y.flatten()*y_scale
    z = z.flatten()
    # first, we need to formulate this as A*t = z
    # where A has shape (x.size, (x_order+1)*(y_order+1))
    # and t has shape ((x_order+1)*(y_order+1), )
    A = numpy.empty((x.size, (x_order+1)*(y_order+1)), dtype=numpy.float64)
    # noinspection PyTypeChecker
    for i, index in enumerate(numpy.ndindex((x_order+1, y_order+1))):
        A[:, i] = numpy.power(x, index[0])*numpy.power(y, index[1])
    # perform least squares fit
    sol, residuals, rank, sing_values = lstsq(A, z, cond=rcond)
    if isinstance(residuals, (numpy.ndarray, numpy.number)):
        residuals /= float(x.size)
    sol = numpy.power(x_scale, numpy.arange(x_order+1))[:, numpy.newaxis] * \
          numpy.reshape(sol, (x_order+1, y_order+1)) * \
          numpy.power(y_scale, numpy.arange(y_order+1))
    return sol, residuals, rank, sing_values


def get_im_physical_coords(array, grid, image_data, direction):
    """
    Converts one dimension of "pixel" image (row or column) coordinates to
    "physical" image (range or azimuth in meters) coordinates, for use in the
    various two-variable sicd polynomials.

    Parameters
    ----------
    array : numpy.array|float|int
        either row or col coordinate component
    grid : sarpy.io.complex.sicd_elements.Grid.GridType
    image_data : sarpy.io.complex.sicd_elements.ImageData.ImageDataType
    direction : str
        one of 'Row' or 'Col' (case insensitive) to determine which
    Returns
    -------
    numpy.array|float
    """

    if direction.upper() == 'ROW':
        return (array - image_data.SCPPixel.Row + image_data.FirstRow)*grid.Row.SS
    elif direction.upper() == 'COL':
        return (array - image_data.SCPPixel.Col + image_data.FirstCol)*grid.Col.SS
    else:
        raise ValueError('Unrecognized direction {}'.format(direction))


def fit_time_coa_polynomial(inca, image_data, grid, dop_rate_scaled_coeffs, poly_order=2):
    """

    Parameters
    ----------
    inca : sarpy.io.complex.sicd_elements.RMA.INCAType
    image_data : sarpy.io.complex.sicd_elements.ImageData.ImageDataType
    grid : sarpy.io.complex.sicd_elements.Grid.GridType
    dop_rate_scaled_coeffs : numpy.ndarray
        the dop rate polynomial relative to physical coordinates - this a
        common construct in converting metadata for csk/sentinel/radarsat
    poly_order : int
        the degree of the polynomial to fit.
    Returns
    -------
    Poly2DType
    """

    grid_samples = poly_order + 8
    coords_az = get_im_physical_coords(
        numpy.linspace(0, image_data.NumCols - 1, grid_samples, dtype='float64'), grid, image_data, 'col')
    coords_rg = get_im_physical_coords(
        numpy.linspace(0, image_data.NumRows - 1, grid_samples, dtype='float64'), grid, image_data, 'row')
    coords_az_2d, coords_rg_2d = numpy.meshgrid(coords_az, coords_rg)
    time_ca_sampled = inca.TimeCAPoly(coords_az_2d)
    dop_centroid_sampled = inca.DopCentroidPoly(coords_rg_2d, coords_az_2d)
    doppler_rate_sampled = polynomial.polyval(coords_rg_2d, dop_rate_scaled_coeffs)
    time_coa_sampled = time_ca_sampled + dop_centroid_sampled / doppler_rate_sampled
    coefs, residuals, rank, sing_values = two_dim_poly_fit(
        coords_rg_2d, coords_az_2d, time_coa_sampled,
        x_order=poly_order, y_order=poly_order, x_scale=1e-3, y_scale=1e-3, rcond=1e-40)
    logger.info(
        'The time_coa_fit details:\n\t'
        'root mean square residuals = {}\n\t'
        'rank = {}\n\t'
        'singular values = {}'.format(residuals, rank, sing_values))
    return Poly2DType(Coefs=coefs)


def fit_position_xvalidation(time_array, position_array, velocity_array, max_degree=5):
    """
    Empirically fit the polynomials for the X, Y, Z ECF position array, using cross
    validation with the velocity array to determine the best fit degree up to a
    given maximum degree.

    Parameters
    ----------
    time_array : numpy.ndarray
    position_array : numpy.ndarray
    velocity_array : numpy.ndarray
    max_degree : int

    Returns
    -------
    (numpy.ndarray, numpy.ndarray, numpy.ndarray,)
        The X, Y, Z polynomial coefficients.
    """

    if not isinstance(time_array, numpy.ndarray) or \
            not isinstance(position_array, numpy.ndarray) or \
            not isinstance(velocity_array, numpy.ndarray):
        raise TypeError('time_array, position_array, and velocity_array must be numpy.ndarray instances.')

    if time_array.ndim != 1 and time_array.size > 1:
        raise ValueError('time_array must be one-dimensional with at least 2 elements.')

    if position_array.shape != velocity_array.shape:
        raise ValueError('position_array and velocity_array must have the same shape.')

    if position_array.shape[0] != time_array.size:
        raise ValueError('The first dimension of position_array must be the same size as time_array.')

    if position_array.shape[1] != 3:
        raise ValueError('The second dimension of position array must have size 3, '
                         'representing X, Y, Z ECF coordinates.')

    max_degree = int(max_degree)
    if max_degree < 1:
        raise ValueError('max_degree must be at least 1.')
    if max_degree > 10:
        logger.warning('max_degree greater than 10 for polynomial fitting may lead\n\t'
                        'to poorly conditioned (i.e. badly behaved) fit.')

    deg = 1
    prev_vel_error = numpy.inf
    P_x, P_y, P_z = None, None, None
    while deg <= min(max_degree, position_array.shape[0]-1):
        # fit position
        P_x = polynomial.polyfit(time_array, position_array[:, 0], deg=deg)
        P_y = polynomial.polyfit(time_array, position_array[:, 1], deg=deg)
        P_z = polynomial.polyfit(time_array, position_array[:, 2], deg=deg)
        # extract estimated velocities
        vel_est = numpy.hstack(
            (polynomial.polyval(time_array, polynomial.polyder(P_x))[:, numpy.newaxis],
             polynomial.polyval(time_array, polynomial.polyder(P_y))[:, numpy.newaxis],
             polynomial.polyval(time_array, polynomial.polyder(P_z))[:, numpy.newaxis]))
        # check our velocity error
        vel_err = vel_est - velocity_array
        cur_vel_error = numpy.sum((vel_err * vel_err))
        deg += 1
        # stop if the error is not smaller than at the previous step
        if cur_vel_error >= prev_vel_error:
            break
    return P_x, P_y, P_z


def sicd_reader_iterator(reader, partitions=None, polarization=None, band=None):
    """
    Provides an iterator over a collection of partitions (tuple of tuple of integer
    indices for the reader) for a sicd type reader object.

    Parameters
    ----------
    reader : SICDTypeReader
    partitions : None|tuple
        The partitions collection. If None, then partitioning from
        `reader.get_sicd_partitions()` will be used.
    polarization : None|str
        The polarization string to match.
    band : None|str
        The band to match.

    Returns
    -------
    Iterator[tuple]
        Yields the partition index, the sicd reader index, and the sicd structure.
    """

    def sicd_match():
        match = True
        if band is not None:
            match &= (this_sicd.get_transmit_band_name() == band)
        if polarization is not None:
            match &= (this_sicd.get_processed_polarization() == polarization)
        return match

    if not isinstance(reader, SICDTypeReader):
        raise TypeError('reader must be an instance of SICDTypeReader. Got type {}'.format(type(reader)))
    if reader.reader_type != "SICD":
        raise ValueError('The provided reader must be of SICD type.')

    if partitions is None:
        # noinspection PyUnresolvedReferences
        partitions = reader.get_sicd_partitions()
    # noinspection PyUnresolvedReferences
    the_sicds = reader.get_sicds_as_tuple()
    for this_partition, entry in enumerate(partitions):
        for this_index in entry:
            this_sicd = the_sicds[this_index]
            if sicd_match():
                yield this_partition, this_index, this_sicd


def get_physical_coordinates(the_sicd, row_value, col_value):
    """
    Transform from image coordinates to physical coordinates, for polynomial evaluation.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType
    row_value : int|float|numpy.ndarray
    col_value : int|float|numpy.ndarray

    Returns
    -------
    tuple
        The row and colummn physical coordinates
    """

    return get_im_physical_coords(row_value, the_sicd.Grid, the_sicd.ImageData, 'row'), \
           get_im_physical_coords(col_value, the_sicd.Grid, the_sicd.ImageData, 'col')



###################
# helper functions

def get_fetch_block_size(start_element, stop_element, block_size_in_bytes, bands=1):
    """
    Gets the appropriate block size, given fetch parameters and constraints.

    Note: this is a helper function, and no checking of argument validity will
    be performed.

    Parameters
    ----------
    start_element : int
    stop_element : int
    block_size_in_bytes : int
    bands : int

    Returns
    -------
    int
    """

    if stop_element == start_element:
        return None
    if block_size_in_bytes is None:
        return None

    full_size = float(abs(stop_element - start_element))
    return max(1, int(numpy.ceil(block_size_in_bytes / float(bands*8*full_size))))


def extract_blocks(the_range, index_block_size):
    # type: (Tuple[int, int, int], int) -> (List[Tuple[int, int, int]], List[Tuple[int, int]])
    """
    Convert the single range definition into a series of range definitions in
    keeping with fetching of the appropriate block sizes.

    Note: this is a helper function, and no checking of argument validity will
    be performed.

    Parameters
    ----------
    the_range : Tuple[int, int, int]
        The input (off processing axis) range.
    index_block_size : None|int|float
        The size of blocks (number of indices).

    Returns
    -------
    List[Tuple[int, int, int]], List[Tuple[int, int]]
        The sequence of range definitions `(start index, stop index, step)`
        relative to the overall image, and the sequence of start/stop indices
        for positioning of the given range relative to the original range.
    """

    entries = numpy.arange(the_range[0], the_range[1], the_range[2], dtype=numpy.int64)
    if index_block_size is None:
        return [the_range, ], [(0, entries.size), ]

    # how many blocks?
    block_count = int(numpy.ceil(entries.size/float(index_block_size)))
    if index_block_size == 1:
        return [the_range, ], [(0, entries.size), ]

    # workspace for what the blocks are
    out1 = []
    out2 = []
    start_ind = 0
    for i in range(block_count):
        end_ind = start_ind+index_block_size
        if end_ind < entries.size:
            block1 = (int(entries[start_ind]), int(entries[end_ind]), the_range[2])
            block2 = (start_ind, end_ind)
        else:
            block1 = (int(entries[start_ind]), the_range[1], the_range[2])
            block2 = (start_ind, entries.size)
        out1.append(block1)
        out2.append(block2)
        start_ind = end_ind
    return out1, out2


def get_data_mean_magnitude(bounds, reader, index, block_size_in_bytes):
    """
    Gets the mean magnitude in the region defined by bounds.

    Note: this is a helper function, and no checking of argument validity will
    be performed.

    Parameters
    ----------
    bounds : numpy.ndarray|tuple|list
        Of the form `(row_start, row_end, col_start, col_end)`.
    reader : SICDTypeReader
        The data reader.
    index : int
        The reader index to use.
    block_size_in_bytes : int|float
        The block size in bytes.

    Returns
    -------
    float
    """

    # Extract the mean of the data magnitude - for global remap usage
    logger.info(
        'Calculating mean over the block ({}:{}, {}:{}), this may be time consuming'.format(*bounds))
    mean_block_size = get_fetch_block_size(bounds[0], bounds[1], block_size_in_bytes)
    mean_column_blocks, _ = extract_blocks((bounds[2], bounds[3], 1), mean_block_size)
    mean_total = 0.0
    mean_count = 0
    for this_column_range in mean_column_blocks:
        data = numpy.abs(reader[
                         bounds[0]:bounds[1],
                         this_column_range[0]:this_column_range[1],
                         index])
        mask = (data > 0) & numpy.isfinite(data)
        mean_total += numpy.sum(data[mask])
        mean_count += numpy.sum(mask)
    return float(mean_total / mean_count)


def stats_calculation(data, percentile=None):
    """
    Calculate the statistics for input into the nrl remap.

    Parameters
    ----------
    data : numpy.ndarray
        The amplitude array, assumed real valued an finite.
    percentile : None|float|int
        Which percentile to calculate

    Returns
    -------
    tuple
        Of the form `(`minimum, maximum)` or `(minimum, maximum, `percentile` percentile)
    """

    if percentile is None:
        return numpy.min(data), numpy.max(data)
    else:
        return numpy.min(data), numpy.max(data), scoreatpercentile(data, percentile)


def get_data_extrema(bounds, reader, index, block_size_in_bytes, percentile=None):
    """
    Gets the minimum and maximum magnitude in the region defined by bounds,
    optionally, get as **estimate** for the `percentage` percentile point.

    Note: this is a helper function, and no checking of argument validity will
    be performed.

    Parameters
    ----------
    bounds : numpy.ndarray|tuple|list
        Of the form `(row_start, row_end, col_start, col_end)`.
    reader : SICDTypeReader
        The data reader.
    index : int
        The reader index to use.
    block_size_in_bytes : int|float
        The block size in bytes.
    percentile : None|int|float
        The optional percentile to estimate.

    Returns
    -------
    Tuple
        The minimum (finite) observed, maximum (finite) observed, and optionally,
        the desired (of finite) percentile.

        The minimum, maximum, and percentile values will be `None` exactly when the reader
        contains no finite data.
    """

    min_value = None
    max_value = None
    percent = None

    # Extract the mean of the data magnitude - for global remap usage
    logger.info(
        'Calculating extrema over the block ({}:{}, {}:{}), this may be time consuming'.format(*bounds))
    mean_block_size = get_fetch_block_size(bounds[0], bounds[1], block_size_in_bytes)
    mean_column_blocks, _ = extract_blocks((bounds[2], bounds[3], 1), mean_block_size)
    for this_column_range in mean_column_blocks:
        data = numpy.abs(reader[
                         bounds[0]:bounds[1],
                         this_column_range[0]:this_column_range[1],
                         index])
        mask = numpy.isfinite(data)
        if numpy.any(mask):
            temp_values = stats_calculation(data[mask], percentile=percentile)

            min_value = temp_values[0] if min_value is None else min(min_value, temp_values[0])
            max_value = temp_values[1] if max_value is None else max(max_value, temp_values[1])
            percent = temp_values[2] if percentile is not None else None

    if percentile is None:
        return min_value, max_value
    else:
        return min_value, max_value, percent



"""
Functionality for reading NISAR data into a SICD model.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
import os
from collections import OrderedDict
from typing import Tuple, Dict

import numpy
from numpy.polynomial import polynomial
from scipy.constants import speed_of_light

from sarpy.compliance import bytes_to_string
from sarpy.io.complex.base import SICDTypeReader, H5Chipper, h5py, is_hdf5
from sarpy.io.complex.sicd_elements.blocks import Poly2DType
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_elements.CollectionInfo import CollectionInfoType, RadarModeType
from sarpy.io.complex.sicd_elements.ImageCreation import ImageCreationType
from sarpy.io.complex.sicd_elements.RadarCollection import RadarCollectionType, \
    ChanParametersType, TxStepType
from sarpy.io.complex.sicd_elements.ImageData import ImageDataType
from sarpy.io.complex.sicd_elements.GeoData import GeoDataType, SCPType
from sarpy.io.complex.sicd_elements.SCPCOA import SCPCOAType
from sarpy.io.complex.sicd_elements.Position import PositionType, XYZPolyType
from sarpy.io.complex.sicd_elements.Grid import GridType, DirParamType, WgtTypeType
from sarpy.io.complex.sicd_elements.Timeline import TimelineType, IPPSetType
from sarpy.io.complex.sicd_elements.ImageFormation import ImageFormationType, \
    RcvChanProcType
from sarpy.io.complex.sicd_elements.RMA import RMAType, INCAType
from sarpy.io.complex.sicd_elements.Radiometric import RadiometricType, NoiseLevelType_
from sarpy.geometry import point_projection
from sarpy.io.general.base import BaseReader, SarpyIOError
from sarpy.io.general.utils import get_seconds, parse_timestring, is_file_like
from sarpy.io.complex.utils import fit_position_xvalidation, two_dim_poly_fit

logger = logging.getLogger(__name__)

########
# base expected functionality for a module with an implemented Reader


def is_a(file_name):
    """
    Tests whether a given file_name corresponds to a NISAR file. Returns a reader instance, if so.

    Parameters
    ----------
    file_name : str|BinaryIO
        the file_name to check

    Returns
    -------
    NISARReader|None
        `NISARReader` instance if NISAR file, `None` otherwise
    """

    if is_file_like(file_name):
        return None

    if not is_hdf5(file_name):
        return None

    if h5py is None:
        return None

    try:
        nisar_details = NISARDetails(file_name)
        logger.info('File {} is determined to be a NISAR file.'.format(file_name))
        return NISARReader(nisar_details)
    except (ImportError, SarpyIOError):
        return None


###########
# parser and interpreter for hdf5 attributes

def _stringify(val):
    """
    Decode the value as necessary, for hdf5 string support issues.

    Parameters
    ----------
    val : str|bytes

    Returns
    -------
    str
    """

    return bytes_to_string(val).strip()


def _get_ref_time(str_in):
    """
    Extract the given reference time.

    Parameters
    ----------
    str_in : str|bytes

    Returns
    -------
    numpy.datetime64
    """

    str_in = bytes_to_string(str_in)
    prefix = 'seconds since '
    if not str_in.startswith(prefix):
        raise ValueError('Got unexpected reference time string - {}'.format(str_in))
    return parse_timestring(str_in[len(prefix):], precision='ns')


def _get_string_list(array):
    return [bytes_to_string(el) for el in array]


class NISARDetails(object):
    """
    Parses and converts the Cosmo Skymed metadata
    """

    __slots__ = ('_file_name', )

    def __init__(self, file_name):
        """

        Parameters
        ----------
        file_name : str
        """

        if h5py is None:
            raise ImportError("Can't read NISAR files, because the h5py dependency is missing.")

        if not os.path.isfile(file_name):
            raise SarpyIOError('Path {} is not a file'.format(file_name))

        with h5py.File(file_name, 'r') as hf:
            # noinspection PyBroadException
            try:
                # noinspection PyUnusedLocal
                gp = hf['/science/LSAR/SLC']
            except Exception as e:
                raise SarpyIOError('Got an error when reading required path /science/LSAR/SLC\n\t{}'.format(e))

        self._file_name = file_name

    @property
    def file_name(self):
        """
        str: the file name
        """

        return self._file_name

    @staticmethod
    def _get_frequency_list(hf):
        """
        Gets the list of frequencies.

        Parameters
        ----------
        hf : h5py.File

        Returns
        -------
        numpy.ndarray
        """

        return _get_string_list(hf['/science/LSAR/identification/listOfFrequencies'][:])

    @staticmethod
    def _get_collection_times(hf):
        """
        Gets the collection start and end times, and inferred duration.

        Parameters
        ----------
        hf : h5py.File
            The h5py File object.

        Returns
        -------
        (numpy.datetime64, numpy.datetime64, float)
            Start and end times and duration
        """

        start = parse_timestring(_stringify(hf['/science/LSAR/identification/zeroDopplerStartTime'][()]), precision='ns')
        end = parse_timestring(_stringify(hf['/science/LSAR/identification/zeroDopplerEndTime'][()]), precision='ns')
        duration = get_seconds(end, start, precision='ns')
        return start, end, duration

    @staticmethod
    def _get_zero_doppler_data(hf, base_sicd):
        """
        Gets zero-doppler parameters.

        Parameters
        ----------
        hf : h5py.File
        base_sicd : SICDType

        Returns
        -------
        (numpy.ndarray, float, numpy.ndarray, numpy.ndarray)
            The azimuth zero-doppler time array, azimuth zero-doppler time spacing,
            grid range array, range zero doppler time array.
        """

        gp = hf['/science/LSAR/SLC/swaths']
        ds = gp['zeroDopplerTime']
        ref_time = _get_ref_time(ds.attrs['units'])
        zd_time = ds[:] + get_seconds(ref_time, base_sicd.Timeline.CollectStart, precision='ns')
        ss_az_s = gp['zeroDopplerTimeSpacing'][()]

        if base_sicd.SCPCOA.SideOfTrack == 'L':
            zd_time = zd_time[::-1]
            ss_az_s *= -1

        gp = hf['/science/LSAR/SLC/metadata/processingInformation/parameters']
        grid_r = gp['slantRange'][:]
        ds = gp['zeroDopplerTime']
        ref_time = _get_ref_time(ds.attrs['units'])
        grid_zd_time = ds[:] + get_seconds(ref_time, base_sicd.Timeline.CollectStart, precision='ns')

        return zd_time, ss_az_s, grid_r, grid_zd_time

    def _get_base_sicd(self, hf):
        """
        Defines the base SICD object, to be refined with further details.

        Returns
        -------
        SICDType
        """

        def get_collection_info():
            # type: () -> CollectionInfoType
            gp = hf['/science/LSAR/identification']
            return CollectionInfoType(
                CollectorName=_stringify(hf.attrs['mission_name']),
                CoreName='{0:07d}_{1:s}'.format(gp['absoluteOrbitNumber'][()],
                                                _stringify(gp['trackNumber'][()])),
                CollectType='MONOSTATIC',
                Classification='UNCLASSIFIED',
                RadarMode=RadarModeType(ModeType='STRIPMAP'))

        def get_image_creation():
            # type: () -> ImageCreationType
            application = 'ISCE'
            # noinspection PyBroadException
            try:
                application = '{} {}'.format(
                    application,
                    _stringify(hf['/science/LSAR/SLC/metadata/processingInformation/algorithms/ISCEVersion'][()]))
            except Exception as e:
                logger.info('Failed extracting the application details with error\n\t{}'.format(e))
                pass

            from sarpy.__about__ import __version__
            # TODO: DateTime?
            return ImageCreationType(
                Application=application,
                Site='Unknown',
                Profile='sarpy {}'.format(__version__))

        def get_geo_data():
            # type: () -> GeoDataType
            # seeds a rough SCP for projection usage
            poly_str = _stringify(hf['/science/LSAR/identification/boundingPolygon'][()])
            beg_str = 'POLYGON (('
            if not poly_str.startswith(beg_str):
                raise ValueError('Unexpected polygon string {}'.format(poly_str))
            parts = poly_str[len(beg_str):-2].strip().split(',')
            if len(parts) != 5:
                raise ValueError('Unexpected polygon string parts {}'.format(parts))
            lats_lons = numpy.zeros((4, 2), dtype=numpy.float64)
            for i, part in enumerate(parts[:-1]):
                spart = part.strip().split()
                if len(spart) != 2:
                    raise ValueError('Unexpected polygon string parts {}'.format(parts))
                lats_lons[i, :] = float(spart[1]), float(spart[0])

            llh = numpy.zeros((3, ), dtype=numpy.float64)
            llh[0:2] = numpy.mean(lats_lons, axis=0)
            llh[2] = numpy.mean(hf['/science/LSAR/SLC/metadata/processingInformation/parameters/referenceTerrainHeight'][:])
            return GeoDataType(SCP=SCPType(LLH=llh))

        def get_grid():
            # type: () -> GridType

            # TODO: Future Change Required - JPL states that uniform weighting in data simulated
            #  from UAVSAR is a placeholder, not an accurate description of the data.
            #  At this point, it is not clear what the final weighting description for NISAR
            #  will be.

            gp = hf['/science/LSAR/SLC/metadata/processingInformation/parameters']
            row_wgt = gp['rangeChirpWeighting'][:]
            win_name = 'UNIFORM' if numpy.all(row_wgt == row_wgt[0]) else 'UNKNOWN'
            row = DirParamType(
                Sgn=-1,
                DeltaKCOAPoly=[[0,]],
                WgtFunct=numpy.cast[numpy.float64](row_wgt),
                WgtType=WgtTypeType(WindowName=win_name))

            col_wgt = gp['azimuthChirpWeighting'][:]
            win_name = 'UNIFORM' if numpy.all(col_wgt == col_wgt[0]) else 'UNKNOWN'
            col = DirParamType(
                Sgn=-1,
                KCtr=0,
                WgtFunct=numpy.cast[numpy.float64](col_wgt),
                WgtType=WgtTypeType(WindowName=win_name))

            return GridType(ImagePlane='SLANT', Type='RGZERO', Row=row, Col=col)

        def get_timeline():
            # type: () -> TimelineType
            # NB: IPPEnd must be set, but will be replaced
            return TimelineType(
                CollectStart=collect_start,
                CollectDuration=duration,
                IPP=[IPPSetType(index=0, TStart=0, TEnd=duration, IPPStart=0, IPPEnd=0), ])

        def get_position():
            # type: () -> PositionType
            gp = hf['/science/LSAR/SLC/metadata/orbit']
            ref_time = _get_ref_time(gp['time'].attrs['units'])
            T = gp['time'][:] + get_seconds(ref_time, collect_start, precision='ns')
            Pos = gp['position'][:]
            Vel = gp['velocity'][:]
            P_x, P_y, P_z = fit_position_xvalidation(T, Pos, Vel, max_degree=8)
            return PositionType(ARPPoly=XYZPolyType(X=P_x, Y=P_y, Z=P_z))

        def get_scpcoa():
            # type: () -> SCPCOAType
            # remaining fields set later
            sot = _stringify(hf['/science/LSAR/identification/lookDirection'][()])[0].upper()
            return SCPCOAType(SideOfTrack=sot)

        def get_image_formation():
            # type: () -> ImageFormationType
            return ImageFormationType(
                ImageFormAlgo='RMA',
                TStartProc=0,
                TEndProc=duration,
                STBeamComp='NO',
                ImageBeamComp='SV',
                AzAutofocus='NO',
                RgAutofocus='NO',
                RcvChanProc=RcvChanProcType(NumChanProc=1, PRFScaleFactor=1))

        def get_rma():
            # type: () -> RMAType
            return RMAType(RMAlgoType='OMEGA_K', INCA=INCAType(DopCentroidCOA=True))

        collect_start, collect_end, duration = self._get_collection_times(hf)
        collection_info = get_collection_info()
        image_creation = get_image_creation()
        geo_data = get_geo_data()
        grid = get_grid()
        timeline = get_timeline()
        position = get_position()
        scpcoa = get_scpcoa()
        image_formation = get_image_formation()
        rma = get_rma()

        return SICDType(
            CollectionInfo=collection_info,
            ImageCreation=image_creation,
            GeoData=geo_data,
            Grid=grid,
            Timeline=timeline,
            Position=position,
            SCPCOA=scpcoa,
            ImageFormation=image_formation,
            RMA=rma)

    @staticmethod
    def _get_freq_specific_sicd(gp, base_sicd):
        """
        Gets the frequency specific sicd.

        Parameters
        ----------
        gp : h5py.Group
        base_sicd : SICDType

        Returns
        -------
        (SICDType, numpy.ndarray, list, center_freq)
            frequency dependent sicd, array of polarization names, list of formatted polarizations for sicd,
            the processed center frequency
        """

        def update_grid():
            row_imp_resp_bw = 2*gp['processedRangeBandwidth'][()]/speed_of_light
            t_sicd.Grid.Row.SS = gp['slantRangeSpacing'][()]
            t_sicd.Grid.Row.ImpRespBW = row_imp_resp_bw
            t_sicd.Grid.Row.DeltaK1 = -0.5*row_imp_resp_bw
            t_sicd.Grid.Row.DeltaK2 = -t_sicd.Grid.Row.DeltaK1

        def update_timeline():
            prf = gp['nominalAcquisitionPRF'][()]
            t_sicd.Timeline.IPP[0].IPPEnd = prf*t_sicd.Timeline.CollectDuration
            t_sicd.Timeline.IPP[0].IPPPoly = [0, prf]

        def define_radar_collection():
            tx_rcv_pol_t = []
            tx_pol = []
            for entry in pols:
                tx_rcv_pol_t.append('{}:{}'.format(entry[0], entry[1]))
                if entry[0] not in tx_pol:
                    tx_pol.append(entry[0])
            center_freq_t = gp['acquiredCenterFrequency'][()]
            bw = gp['acquiredRangeBandwidth'][()]
            tx_freq = (center_freq_t - 0.5*bw, center_freq_t + 0.5*bw)
            rcv_chans = [ChanParametersType(TxRcvPolarization=pol) for pol in tx_rcv_pol_t]
            if len(tx_pol) == 1:
                tx_sequence = None
                tx_pol = tx_pol[0]
            else:
                tx_sequence = [TxStepType(WFIndex=j+1, TxPolarization=pol) for j, pol in enumerate(tx_pol)]
                tx_pol = 'SEQUENCE'

            t_sicd.RadarCollection = RadarCollectionType(
                TxFrequency=tx_freq,
                RcvChannels=rcv_chans,
                TxPolarization=tx_pol,
                TxSequence=tx_sequence)
            return tx_rcv_pol_t

        def update_image_formation():
            center_freq_t = gp['processedCenterFrequency'][()]
            bw = gp['processedRangeBandwidth'][()]
            t_sicd.ImageFormation.TxFrequencyProc = (center_freq_t - 0.5*bw, center_freq_t + 0.5*bw)
            return center_freq_t

        pols = _get_string_list(gp['listOfPolarizations'][:])
        t_sicd = base_sicd.copy()

        update_grid()
        update_timeline()
        tx_rcv_pol = define_radar_collection()
        center_freq = update_image_formation()
        return t_sicd, pols, tx_rcv_pol, center_freq

    @staticmethod
    def _get_pol_specific_sicd(hf, ds, base_sicd, pol_name, freq_name, j, pol,
                               r_ca_sampled, zd_time, grid_zd_time, grid_r, doprate_sampled,
                               dopcentroid_sampled, center_freq, ss_az_s, dop_bw, beta0, gamma0, sigma0):
        """
        Gets the frequency/polarization specific sicd.

        Parameters
        ----------
        hf : h5py.File
        ds : h5py.Dataset
        base_sicd : SICDType
        pol_name : str
        freq_name : str
        j : int
        pol : str
        r_ca_sampled : numpy.ndarray
        zd_time : numpy.ndarray
        grid_zd_time : numpy.ndarray
        grid_r : numpy.ndarray
        doprate_sampled : numpy.ndarray
        dopcentroid_sampled : numpy.ndarray
        center_freq : float
        ss_az_s : float
        dop_bw : float

        Returns
        -------
        (SICDType, Tuple[int])
        """

        def define_image_data():
            dtype = ds.dtype.name
            if dtype in ('float32', 'complex64'):
                pixel_type = 'RE32F_IM32F'
            elif dtype == 'int16':
                pixel_type = 'RE16I_IM16I'
            else:
                raise ValueError('Got unhandled dtype {}'.format(dtype))
            t_sicd.ImageData = ImageDataType(
                PixelType=pixel_type,
                NumRows=shape[0],
                NumCols=shape[1],
                FirstRow=0,
                FirstCol=0,
                SCPPixel=[0.5*shape[0], 0.5*shape[1]],
                FullImage=[shape[0], shape[1]])

        def update_image_formation():
            t_sicd.ImageFormation.RcvChanProc.ChanIndices = [j, ]
            t_sicd.ImageFormation.TxRcvPolarizationProc = pol

        def update_inca_and_grid():
            t_sicd.RMA.INCA.R_CA_SCP = r_ca_sampled[t_sicd.ImageData.SCPPixel.Row]
            scp_ca_time = zd_time[t_sicd.ImageData.SCPPixel.Col]

            # compute DRateSFPoly
            # velocity at scp ca time
            vel_ca = t_sicd.Position.ARPPoly.derivative_eval(scp_ca_time, der_order=1)
            # squared magnitude
            vm_ca_sq = numpy.sum(vel_ca*vel_ca)
            # polynomial coefficient for function representing range as a function of range distance from SCP
            r_ca_poly = numpy.array([t_sicd.RMA.INCA.R_CA_SCP, 1], dtype=numpy.float64)
            # closest Doppler rate polynomial to SCP
            min_ind = numpy.argmin(numpy.absolute(grid_zd_time - scp_ca_time))
            # define range coordinate grid
            coords_rg_m = grid_r - t_sicd.RMA.INCA.R_CA_SCP
            # determine dop_rate_poly coordinates
            dop_rate_poly = polynomial.polyfit(coords_rg_m, -doprate_sampled[min_ind, :], 4)  # why fourth order?
            t_sicd.RMA.INCA.FreqZero = center_freq
            t_sicd.RMA.INCA.DRateSFPoly = Poly2DType(Coefs=numpy.reshape(
                -numpy.convolve(dop_rate_poly, r_ca_poly)*speed_of_light/(2*center_freq*vm_ca_sq), (-1, 1)))

            # update Grid.Col parameters
            t_sicd.Grid.Col.SS = numpy.sqrt(vm_ca_sq)*abs(ss_az_s)*t_sicd.RMA.INCA.DRateSFPoly.Coefs[0, 0]
            t_sicd.Grid.Col.ImpRespBW = min(abs(dop_bw*ss_az_s), 1)/t_sicd.Grid.Col.SS
            t_sicd.RMA.INCA.TimeCAPoly = [scp_ca_time, ss_az_s/t_sicd.Grid.Col.SS]

            #TimeCOAPoly/DopCentroidPoly/DeltaKCOAPoly
            coords_az_m = (grid_zd_time - scp_ca_time)*t_sicd.Grid.Col.SS/ss_az_s

            # cerate the 2d grids
            coords_rg_2d_t, coords_az_2d_t = numpy.meshgrid(coords_rg_m, coords_az_m, indexing='xy')

            coefs, residuals, rank, sing_values = two_dim_poly_fit(
                coords_rg_2d_t, coords_az_2d_t, dopcentroid_sampled,
                x_order=3, y_order=3, x_scale=1e-3, y_scale=1e-3, rcond=1e-40)
            logger.info(
                'The dop_centroid_poly fit details:\n\t'
                'root mean square residuals = {}\n\t'
                'rank = {}\n\t'
                'singular values = {}'.format(residuals, rank, sing_values))
            t_sicd.RMA.INCA.DopCentroidPoly = Poly2DType(Coefs=coefs)
            t_sicd.Grid.Col.DeltaKCOAPoly = Poly2DType(Coefs=coefs*ss_az_s/t_sicd.Grid.Col.SS)

            timeca_sampled = numpy.outer(grid_zd_time, numpy.ones((grid_r.size, )))
            time_coa_sampled = timeca_sampled + (dopcentroid_sampled/doprate_sampled)
            coefs, residuals, rank, sing_values = two_dim_poly_fit(
                coords_rg_2d_t, coords_az_2d_t, time_coa_sampled,
                x_order=3, y_order=3, x_scale=1e-3, y_scale=1e-3, rcond=1e-40)
            logger.info(
                'The time_coa_poly fit details:\n\t'
                'root mean square residuals = {}\n\t'
                'rank = {}\n\t'
                'singular values = {}'.format(residuals, rank, sing_values))
            t_sicd.Grid.TimeCOAPoly = Poly2DType(Coefs=coefs)

            return coords_rg_2d_t, coords_az_2d_t

        def define_radiometric():
            def get_poly(ds, name):
                array = ds[:]
                fill = ds.attrs['_FillValue']
                boolc = (array != fill)

                if numpy.any(boolc):
                    array = array[boolc]
                    if numpy.any(array != array[0]):
                        coefs, residuals, rank, sing_values = two_dim_poly_fit(
                            coords_rg_2d[boolc], coords_az_2d[boolc], array,
                            x_order=3, y_order=3, x_scale=1e-3, y_scale=1e-3, rcond=1e-40)
                        logger.info(
                            'The {} fit details:\n\t'
                            'root mean square residuals = {}\n\t'
                            'rank = {}\n\t'
                            'singular values = {}'.format(name, residuals, rank, sing_values))
                    else:
                        # it's constant, so just use a constant polynomial
                        coefs = [[array[0], ], ]
                        logger.info('The {} values are constant'.format(name))
                    return Poly2DType(Coefs=coefs)
                else:
                    logger.warning('No non-trivial values for {} provided.'.format(name))
                    return None

            beta0_poly = get_poly(beta0, 'beta0')
            gamma0_poly = get_poly(gamma0, 'gamma0')
            sigma0_poly = get_poly(sigma0, 'sigma0')

            nesz = hf['/science/LSAR/SLC/metadata/calibrationInformation/frequency{}/{}/nes0'.format(freq_name,
                                                                                                     pol_name)][:]
            noise_samples = nesz - (10 * numpy.log10(sigma0_poly.Coefs[0, 0]))

            coefs, residuals, rank, sing_values = two_dim_poly_fit(
                coords_rg_2d, coords_az_2d, noise_samples,
                x_order=3, y_order=3, x_scale=1e-3, y_scale=1e-3, rcond=1e-40)
            logger.info(
                'The noise_poly fit details:\n\t'
                'root mean square residuals = {}\n\t'
                'rank = {}\n\t'
                'singular values = {}'.format(
                    residuals, rank, sing_values))
            t_sicd.Radiometric = RadiometricType(
                BetaZeroSFPoly=beta0_poly,
                GammaZeroSFPoly=gamma0_poly,
                SigmaZeroSFPoly=sigma0_poly,
                NoiseLevel=NoiseLevelType_(
                    NoiseLevelType='ABSOLUTE', NoisePoly=Poly2DType(Coefs=coefs)))

        def update_geodata():
            ecf = point_projection.image_to_ground(
                [t_sicd.ImageData.SCPPixel.Row, t_sicd.ImageData.SCPPixel.Col], t_sicd)
            t_sicd.GeoData.SCP = SCPType(ECF=ecf)  # LLH will be populated

        t_sicd = base_sicd.copy()
        shape = (int(ds.shape[1]), int(ds.shape[0]))

        define_image_data()
        update_image_formation()
        coords_rg_2d, coords_az_2d = update_inca_and_grid()
        define_radiometric()
        update_geodata()
        t_sicd.derive()
        t_sicd.populate_rniirs(override=False)
        return t_sicd, shape

    def get_sicd_collection(self):
        """
        Get the sicd collection for the bands.

        Returns
        -------
        Tuple[Dict[str, SICDType], Dict[str, str], Tuple[bool, bool, bool]]
            the first entry is a dictionary of the form {band_name: sicd}
            the second entry is of the form {band_name: shape}
            the third entry is the symmetry tuple
        """

        # TODO: check if the hdf already has the sicds defined, and fish them out if so.

        with h5py.File(self.file_name, 'r') as hf:
            # fetch the base shared sicd
            base_sicd = self._get_base_sicd(hf)

            # prepare our output workspace
            out_sicds = OrderedDict()
            shapes = OrderedDict()
            symmetry = (base_sicd.SCPCOA.SideOfTrack == 'L', False, True)

            # fetch the common use data for frequency issues
            collect_start, collect_end, duration = self._get_collection_times(hf)
            zd_time, ss_az_s, grid_r, grid_zd_time = self._get_zero_doppler_data(hf, base_sicd)

            gp = hf['/science/LSAR/SLC/metadata/calibrationInformation/geometry']
            beta0 = gp['beta0']
            gamma0 = gp['gamma0']
            sigma0 = gp['sigma0']

            # formulate the frequency specific sicd information
            freqs = self._get_frequency_list(hf)
            for i, freq in enumerate(freqs):
                gp_name = '/science/LSAR/SLC/swaths/frequency{}'.format(freq)
                gp = hf[gp_name]
                freq_sicd, pols, tx_rcv_pol, center_freq = self._get_freq_specific_sicd(gp, base_sicd)

                # formulate the frequency dependent doppler grid
                # TODO: Future Change Required - processedAzimuthBandwidth acknowledged
                #  by JPL to be wrong in simulated datasets.
                dop_bw = gp['processedAzimuthBandwidth'][()]
                gp2 = hf['/science/LSAR/SLC/metadata/processingInformation/parameters/frequency{}'.format(freq)]
                dopcentroid_sampled = gp2['dopplerCentroid'][:]
                doprate_sampled = gp2['azimuthFMRate'][:]
                r_ca_sampled = gp['slantRange'][:]
                # formulate the frequency/polarization specific sicd information
                for j, pol in enumerate(pols):
                    ds_name = '{}/{}'.format(gp_name, pol)
                    ds = gp[pol]
                    pol_sicd, shape = self._get_pol_specific_sicd(
                        hf, ds, freq_sicd, pol, freq, j, tx_rcv_pol[j],
                        r_ca_sampled, zd_time, grid_zd_time, grid_r,
                        doprate_sampled, dopcentroid_sampled, center_freq,
                        ss_az_s, dop_bw, beta0, gamma0, sigma0)
                    out_sicds[ds_name] = pol_sicd
                    shapes[ds_name] = ds.shape[:2]
        return out_sicds, shapes, symmetry


################
# The NISAR reader

class NISARReader(BaseReader, SICDTypeReader):
    """
    Gets a reader type object for NISAR files
    """

    __slots__ = ('_nisar_details', )

    def __init__(self, nisar_details):
        """

        Parameters
        ----------
        nisar_details : str|NISARDetails
            file name or NISARDetails object
        """

        if isinstance(nisar_details, str):
            nisar_details = NISARDetails(nisar_details)
        if not isinstance(nisar_details, NISARDetails):
            raise TypeError('The input argument for NISARReader must be a '
                            'filename or NISARDetails object')
        self._nisar_details = nisar_details
        sicd_data, shape_dict, symmetry = nisar_details.get_sicd_collection()
        chippers = []
        sicds = []
        for band_name in sicd_data:
            sicds.append(sicd_data[band_name])
            chippers.append(H5Chipper(nisar_details.file_name, band_name, shape_dict[band_name], symmetry))

        SICDTypeReader.__init__(self, tuple(sicds))
        BaseReader.__init__(self, tuple(chippers), reader_type="SICD")
        self._check_sizes()

    @property
    def nisar_details(self):
        # type: () -> NISARDetails
        """
        NISARDetails: The nisar details object.
        """

        return self._nisar_details

    @property
    def file_name(self):
        return self.nisar_details.file_name



"""
The GeoInfo definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from collections import OrderedDict
from xml.etree import ElementTree
from typing import List, Union, Dict

import numpy
from sarpy.io.xml.base import Serializable, SerializableArray, ParametersCollection, \
    find_children
from sarpy.io.xml.descriptors import StringDescriptor, SerializableArrayDescriptor, \
    ParametersDescriptor
from sarpy.io.complex.sicd_elements.blocks import LatLonRestrictionType, LatLonArrayElementType

from .base import DEFAULT_STRICT


class LineType(Serializable):
    _fields = ('EndPoints', )
    _required = ('EndPoints', )
    _collections_tags = {'EndPoints': {'array': True, 'child_tag': 'Endpoint'}}
    # descriptors
    EndPoints = SerializableArrayDescriptor(
        'EndPoints', LatLonArrayElementType, _collections_tags, _required, strict=DEFAULT_STRICT, minimum_length=2,
        docstring='A geographic line (array) with WGS-84 coordinates.'
    )  # type: Union[SerializableArray, List[LatLonArrayElementType]]

    def __init__(self, EndPoints=None, **kwargs):
        """

        Parameters
        ----------
        EndPoints : SerializableArray|List[LatLonArrayElementType]|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.EndPoints = EndPoints
        super(LineType, self).__init__(**kwargs)


class PolygonType(Serializable):
    _fields = ('Polygon', )
    _required = ('Polygon', )
    _collections_tags = {'Polygon': {'array': True, 'child_tag': 'Vertex'}}
    # descriptors
    Polygon = SerializableArrayDescriptor(
        'Polygon', LatLonArrayElementType, _collections_tags, _required, strict=DEFAULT_STRICT, minimum_length=3,
        docstring='A geographic polygon (array) with WGS-84 coordinates.'
    )  # type: Union[SerializableArray, List[LatLonArrayElementType]]

    def __init__(self, Polygon=None, **kwargs):
        """

        Parameters
        ----------
        Polygon: SerializableArray|List[LatLonArrayElementType]|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Polygon = Polygon
        super(PolygonType, self).__init__(**kwargs)


class GeoInfoType(Serializable):
    """
    A geographic feature.
    """

    _fields = ('name', 'Descriptions', 'Point', 'Line', 'Polygon', 'GeoInfo')
    _required = ('name', )
    _set_as_attribute = ('name', )
    _collections_tags = {
        'Descriptions': {'array': False, 'child_tag': 'Desc'},
        'Point': {'array': True, 'child_tag': 'Point'},
        'Line': {'array': True, 'child_tag': 'Line'},
        'Polygon': {'array': True, 'child_tag': 'Polygon'},
        'GeoInfo': {'array': False, 'child_tag': 'GeoInfo'}
    }
    # descriptors
    name = StringDescriptor(
        'name', _required, strict=DEFAULT_STRICT,
        docstring='The name.')  # type: str
    Descriptions = ParametersDescriptor(
        'Descriptions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Descriptions of the geographic feature.')  # type: ParametersCollection
    Point = SerializableArrayDescriptor(
        'Point', LatLonRestrictionType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Geographic points with WGS-84 coordinates.'
    )  # type: Union[SerializableArray, List[LatLonRestrictionType]]
    Line = SerializableArrayDescriptor(
        'Line', LineType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Geographic lines (array) with WGS-84 coordinates.'
    )  # type: Union[SerializableArray, List[LineType]]
    Polygon = SerializableArrayDescriptor(
        'Polygon', PolygonType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Geographic polygons (array) with WGS-84 coordinates.'
    )  # type: Union[SerializableArray, List[PolygonType]]

    def __init__(self, name=None, Descriptions=None, Point=None, Line=None,
                 Polygon=None, GeoInfo=None, **kwargs):
        """

        Parameters
        ----------
        name : str
        Descriptions : ParametersCollection|dict
        Point : SerializableArray|List[LatLonRestrictionType]|numpy.ndarray|list|tuple
        Line : SerializableArray|List[LineType]
        Polygon : SerializableArray|List[PolygonType]
        GeoInfo : Dict[GeoInfoTpe]
        kwargs : dict
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.name = name
        self.Descriptions = Descriptions
        self.Point = Point
        self.Line = Line
        self.Polygon = Polygon

        self._GeoInfo = []
        if GeoInfo is None:
            pass
        elif isinstance(GeoInfo, GeoInfoType):
            self.addGeoInfo(GeoInfo)
        elif isinstance(GeoInfo, (list, tuple)):
            for el in GeoInfo:
                self.addGeoInfo(el)
        else:
            raise ValueError('GeoInfo got unexpected type {}'.format(type(GeoInfo)))
        super(GeoInfoType, self).__init__(**kwargs)

    @property
    def GeoInfo(self):
        """
        List[GeoInfoType]: list of GeoInfo objects.
        """

        return self._GeoInfo

    def getGeoInfo(self, key):
        """
        Get GeoInfo(s) with name attribute == `key`.

        Parameters
        ----------
        key : str

        Returns
        -------
        List[GeoInfoType]
        """

        return [entry for entry in self._GeoInfo if entry.name == key]

    def addGeoInfo(self, value):
        """
        Add the given GeoInfo to the GeoInfo list.

        Parameters
        ----------
        value : GeoInfoType

        Returns
        -------
        None
        """

        if isinstance(value, ElementTree.Element):
            gi_key = self._child_xml_ns_key.get('GeoInfo', self._xml_ns_key)
            value = GeoInfoType.from_node(value, self._xml_ns, ns_key=gi_key)
        elif isinstance(value, dict):
            value = GeoInfoType.from_dict(value)

        if isinstance(value, GeoInfoType):
            self._GeoInfo.append(value)
        else:
            raise TypeError('Trying to set GeoInfo element with unexpected type {}'.format(type(value)))

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        if kwargs is None:
            kwargs = OrderedDict()
        gi_key = cls._child_xml_ns_key.get('GeoInfo', ns_key)
        kwargs['GeoInfo'] = find_children(node, 'GeoInfo', xml_ns, gi_key)
        return super(GeoInfoType, cls).from_node(node, xml_ns, ns_key=ns_key, kwargs=kwargs)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        node = super(GeoInfoType, self).to_node(
            doc, tag, ns_key=ns_key, parent=parent, check_validity=check_validity,
            strict=strict, exclude=exclude+('GeoInfo', ))
        # slap on the GeoInfo children
        if self._GeoInfo is not None and len(self._GeoInfo) > 0:
            for entry in self._GeoInfo:
                entry.to_node(doc, tag, ns_key=ns_key, parent=node, strict=strict)
        return node

    def to_dict(self, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        out = super(GeoInfoType, self).to_dict(
            check_validity=check_validity, strict=strict, exclude=exclude+('GeoInfo', ))
        # slap on the GeoInfo children
        if self.GeoInfo is not None and len(self.GeoInfo) > 0:
            out['GeoInfo'] = [entry.to_dict(check_validity=check_validity, strict=strict) for entry in self._GeoInfo]
        return out



"""
Utilities for parsing slice input.
"""

__classification__ = "UNCLASSIFIED"
__author__ = 'Thomas McCullough'


def validate_slice_int(the_int, bound, include=True):
    """
    Ensure that the given integer makes sense as a slice entry, and move to
    a normalized form.

    Parameters
    ----------
    the_int : int
    bound : int
    include : bool

    Returns
    -------
    int
    """

    if not isinstance(bound, int) or bound <= 0:
        raise TypeError('bound must be a positive integer.')
    if include:
        if the_int <= -bound or the_int >= bound:
            raise ValueError('Slice argument {} does not fit with bound {}'.format(the_int, bound))
    else:
        if the_int <= -bound or the_int > bound:
            raise ValueError('Slice argument {} does not fit with bound {}'.format(the_int, bound))
    if the_int < 0:
        return the_int + bound
    return the_int


def validate_slice(the_slice, bound):
    """
    Parse a slice into a normalized form.

    Parameters
    ----------
    the_slice : slice
    bound : int

    Returns
    -------
    slice
    """

    if not isinstance(the_slice, slice):
        raise TypeError('the_slice must be a of type slice, got type {}.'.format(type(the_slice)))
    if not isinstance(bound, int) or bound <= 0:
        raise TypeError('bound must be a positive integer.')

    t_start = the_slice.start
    t_stop = the_slice.stop
    t_step = 1 if the_slice.step is None else the_slice.step
    if t_start is None and t_stop is None:
        t_start, t_stop = 0, bound
    elif t_start is None:
        t_stop = validate_slice_int(t_stop, bound, include=False)
        t_start = 0 if t_stop >= 0 else bound - 1
    elif t_stop is None:
        t_start = validate_slice_int(t_start, bound)
        t_stop = -1 if t_step < 0 else bound
    else:
        t_start = validate_slice_int(t_start, bound)
        t_stop = validate_slice_int(t_stop, bound, include=False)
    if (t_step < 0 and t_start < t_stop) or (t_step > 0 and t_start > t_stop):
        raise ValueError(
            'The slice values start={}, stop={}, step={} are not viable'.format(t_start, t_stop, t_step))
    return slice(t_start, t_stop, t_step)
